package de.lmu.ifi.pms.cep.dura.compiler.durac.collections;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.*;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.TreeIterator;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCQueryCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraBasicTypesMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraExternalModifierMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraStreamTypesMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraVisibilityModifier;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnknownStreamException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraVisibilityException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStream;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DefaultDefinition;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBoundConstraint;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.ActionStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.EventStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.StateStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchemaTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


/**
 * Contains the (implicit and explicit) schema for event, stateful object and
 * action streams.
 * 
 * @author hausmann
 * 
 */
public class StreamDefinitions {
	/** a mapping from stream types and full qualified stream names to actual TSA streams */ 
	private final HashMap<TemporalStreamSchemaTypeID, HashMap<String, Definition<TSA_TemporalStream>>> stream_schema = new HashMap<TemporalStreamSchemaTypeID, HashMap<String, Definition<TSA_TemporalStream>>>();
	
	private StreamDefinitions() {
		for (TemporalStreamSchemaTypeID type : TemporalStreamSchemaTypeID.values()) {
			this.stream_schema.put(type, new HashMap<String, Definition<TSA_TemporalStream>>());
		}
	};
	
	public StreamDefinitions(final DuraTree definitions, final Preamble preamble, final TypeDefinitions types) {
		this();
		
		assert(definitions.getType() == DEFINITIONS);
		
		HashMap<TemporalStreamSchemaTypeID, HashMap<String, FixedArrayNavigableSet<DTSA_BufferType>>> implicitModifiers = new HashMap<TemporalStreamSchemaTypeID, HashMap<String,FixedArrayNavigableSet<DTSA_BufferType>>>();

		for (DuraTree definition : new TreeIterator(definitions, EVENT, STATEFUL_OBJECT, ACTION)) {
			DuraTree modifier = definition.getChild(0);
			DuraTree schema = definition.getChild(1);
			DuraTree supplement = schema.getChild(1);
			
			TemporalStreamSchemaTypeID streamType = DuraStreamTypesMapping.fromToken(definition);
			FixedArrayNavigableSet<DTSA_BufferType> modifiers = FixedArrayNavigableSet.merge(new FixedArrayNavigableSet<DTSA_BufferType>(DuraExternalModifierMapping.fromToken(modifier)), new FixedArrayNavigableSet<DTSA_BufferType>(DTSA_BufferType.WORKING));
			DuraVisibilityModifier visibility = DuraVisibilityModifier.toVisibilityModifier(modifier);

			String simple_name = schema.getChild(0).getText();
			String full_name = TSA_SyntaxPatterns.bufferFullName(preamble.package_name, simple_name);

			//for normally implicit definitions just memorize the modifiers			
			if (simple_name.contains(labelSeparator)) {
				if (! implicitModifiers.containsKey(streamType)) {
					implicitModifiers.put(streamType, new HashMap<String, FixedArrayNavigableSet<DTSA_BufferType>>());
				}

				if (implicitModifiers.get(streamType).containsKey(full_name)) {
					throw new DuraCompilationException(String.format("Multiple definitions for stream %s found.", full_name), schema);
				}
				
				implicitModifiers.get(streamType).put(full_name, modifiers);
				
				continue;
			}
			
			ArrayList<ConditionFormula> tupleConstraints = new ArrayList<ConditionFormula>();
			ArrayList<ConditionFormula> relationConstraints = new ArrayList<ConditionFormula>();
			
			final HashMap<String, Type> explicitAttributes = getExplicitStreamAttributes(schema, types, preamble, streamType);
	
			final HashMap<String, Type> allAttributes = (HashMap<String, Type>) explicitAttributes.clone();
			addImplicitStreamAttributes(allAttributes, streamType);


			for (int j=0; j<supplement.getChildCount(); j++) {
				DuraTree child = supplement.getChild(j);

				switch (child.getType()) {
				case TUPLE_CONSTRAINTS:
					tupleConstraints.addAll(compileConstraints(child, allAttributes));
					
					break;
				
				case RELATION_CONSTRAINTS:
					relationConstraints.addAll(compileConstraints(child, allAttributes));
					
					break;
				}
			}
			
			if (streamType==TemporalStreamSchemaTypeID.STATE && modifiers.contains(DTSA_BufferType.INPUT)) {
				this.addStatefulObjectInputStream(full_name, definition, explicitAttributes, visibility, tupleConstraints, relationConstraints);
			}
			
			
			if (! DuraExternalModifierMapping.containsStaticModifier(modifier)) {
				addImplicitIdStreamBound(relationConstraints, streamType);
				addImplicitReceptionTimeStreamBound(relationConstraints, streamType);
			} else if (streamType != TemporalStreamSchemaTypeID.STATE) {
				throw new DuraCompilationException("Only stateful objects can be declared static.", modifier);
			}
			
			if (modifiers.contains(DTSA_BufferType.INPUT)) {
				addDefaultConstraintsForStreamBounds(relationConstraints, streamType);
			}

			TemporalStreamSchema streamSchema;
			Definition<TSA_TemporalStream> streamDefinition;
			
			try{
				switch (streamType) {
				case GENERAL:
				case EVENT:
					streamSchema = new EventStreamSchema(new FixedArrayNavigableMap<String, Type>(allAttributes), new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
					streamDefinition = new Definition<TSA_TemporalStream>(full_name, visibility, new TSA_TemporalStream(full_name, "", streamSchema, modifiers, definition.getPositionDescription()));
					
					break;
					
				case STATE:
					streamSchema = new StateStreamSchema(new FixedArrayNavigableMap<String, Type>(allAttributes), new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
					streamDefinition = new Definition<TSA_TemporalStream>(full_name, visibility, new TSA_TemporalStream(full_name, "", streamSchema, modifiers, definition.getPositionDescription()));
					
					this.addImpilictStatefulObjectStreams(definition, streamDefinition, modifiers);
					
					continue;
	
				case ACTION:
					streamSchema = new ActionStreamSchema(new FixedArrayNavigableMap<String, Type>(allAttributes), new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
					streamDefinition = new Definition<TSA_TemporalStream>(full_name, visibility, new TSA_TemporalStream(full_name, "", streamSchema, modifiers, definition.getPositionDescription()));
					
					this.addImpilictActionStreams(definition, streamDefinition);
					
					continue;
					
				default:
					throw new IllegalArgumentException();
				}
			} catch (TSA_SchemaException e) {
				throw new DuraCompilationException(e, definition);
			}

			if (this.stream_schema.get(streamType).containsKey(full_name)) {
				throw new DuraCompilationException(String.format("multiple definitions of stream %s found", full_name), definition);
			}
			
			this.stream_schema.get(streamType).put(full_name, streamDefinition);
		}
		
		//adapt the modifier of normally implicit definitions
		for (TemporalStreamSchemaTypeID streamType : implicitModifiers.keySet()) {
			for (String name : implicitModifiers.get(streamType).keySet()) {
				if (! this.stream_schema.get(streamType).containsKey(name)) {
					throw new DuraCompilationException(String.format("Invalid stream name: %s. User defined streams must not contain a dollar sign ($).", name), definitions);
				}
				
				Definition<TSA_TemporalStream> stream = this.stream_schema.get(streamType).get(name);
				TSA_TemporalStream updatedStream = new TSA_TemporalStream(name, stream.definition.description, stream.definition.schema, implicitModifiers.get(streamType).get(name), stream.definition.origin);
				
				this.stream_schema.get(streamType).put(name, new Definition<TSA_TemporalStream>(stream.name, stream.visibility, updatedStream));
			}
		}
	}
	
	public FixedArray<TSA_TemporalStream> getStreams() {
		ArrayList<TSA_TemporalStream> result = new ArrayList<TSA_TemporalStream>();
		
		for (TemporalStreamSchemaTypeID streamType : this.stream_schema.keySet()) {
			for (Definition<TSA_TemporalStream> definition : this.stream_schema.get(streamType).values()) {
				result.add(definition.definition);
			}
		}
		
		return new FixedArray<TSA_TemporalStream>(result);
	}

	public Set<String> getStreamNames(final TemporalStreamSchemaTypeID streamType) {
		return this.stream_schema.get(streamType).keySet();
	}
	
	public TSA_TemporalStream getStreamByNameUngarded(final TemporalStreamSchemaTypeID streamType, final String streamFullName) throws DuraUnspecificCompliationException {
		assert TSA_SyntaxPatterns.isValidBufferFullName(streamFullName);
		
		return this.getStreamDefinition(streamType, null, streamFullName).definition;
	}
	
	
	public TSA_TemporalStream getStreamByName(final TemporalStreamSchemaTypeID streamType, final Preamble preamble, final String streamName) throws DuraUnknownStreamException, DuraVisibilityException {
		Definition<TSA_TemporalStream> definition = getStreamDefinition(streamType, preamble, streamName);
		
		if (definition.visibility==DuraVisibilityModifier.PUBLIC || definition.name.startsWith(preamble.package_name)) {
			return definition.definition;
		} else {
			throw new DuraVisibilityException(definition.name);
		}
	}
	
	private Definition<TSA_TemporalStream> getStreamDefinition(final TemporalStreamSchemaTypeID streamType, final Preamble preamble, final String streamName) throws DuraUnknownStreamException {
		HashMap<String, Definition<TSA_TemporalStream>> definitions = this.stream_schema.get(streamType);

		String streamPackage = TSA_SyntaxPatterns.bufferFullName(streamName)[0];
		
		if (!streamPackage.equals("") || preamble==null) {
			if (! definitions.containsKey(streamName)) {
				throw new DuraUnknownStreamException(streamName);
			}
			
			return definitions.get(streamName);
		} else if (definitions.containsKey(TSA_SyntaxPatterns.bufferFullName(preamble.package_name, streamName))){
			String full_name = TSA_SyntaxPatterns.bufferFullName(preamble.package_name, streamName);
			
			return this.stream_schema.get(streamType).get(full_name);
		} else if (preamble.imports.containsKey(streamName)) {
			String full_name = preamble.imports.get(streamName);
			
			return this.stream_schema.get(streamType).get(full_name);
		}
		
		throw new DuraUnknownStreamException(streamName);
	}
	
	public boolean containsStream(final TemporalStreamSchemaTypeID streamType, final Preamble preamble, final String streamName) {
		try {
			assert(streamType == TemporalStreamSchemaTypeID.EVENT);
			
			this.getStreamByName(streamType, preamble, streamName);
		} catch (DuraUnspecificCompliationException e) {
			return false;
		}
		
		return true;
	}

	
	public static StreamDefinitions join(final Collection<StreamDefinitions> schemata) {
		StreamDefinitions result = new StreamDefinitions();
		
		for (StreamDefinitions schema : schemata) {
			assert(result.disjoint(schema));
			
			for (TemporalStreamSchemaTypeID type : schema.stream_schema.keySet()) {
				result.stream_schema.get(type).putAll(schema.stream_schema.get(type));
			}
		}
		
		return result;
	}
	
	private boolean disjoint(final StreamDefinitions schema) {
		for (TemporalStreamSchemaTypeID type : schema.stream_schema.keySet()) {
			for (String name : schema.stream_schema.get(type).keySet()) {
				if (this.stream_schema.get(type).containsKey(name)) {
					return false;
				}
			}
		}
				
		return true;
	}
	 
	private static HashMap<String, Type> getExplicitStreamAttributes(final DuraTree schema, final TypeDefinitions typeDefinitions, final Preamble preamble, final TemporalStreamSchemaTypeID streamType) {
		assert(schema.getType() == SCHEMA);

		DuraTree streamAttributes = schema.getChild(0);
		HashMap<String, Type> attributes = new HashMap<String, Type>();
		
		getExplicitStreamAttributes(streamAttributes, new Stack<String>(), attributes, typeDefinitions, preamble);
				
		return attributes;
	}
	
	private static void getExplicitStreamAttributes(final DuraTree term, final Stack<String> path, final HashMap<String, Type> attributes, final TypeDefinitions typeDefinitions, final Preamble preamble) {
		assert(term.getType() == TERM);

		for (int i=0; i<term.getChildCount(); i++) {
			DuraTree child = term.getChild(i);

			switch (child.getType()) {
			case TERM:
				path.push(child.getText());

				getExplicitStreamAttributes(child, path, attributes, typeDefinitions, preamble);

				path.pop();

				break;
			
			case BASIC_TYPE:
				attributes.put(DuraCQueryCompiler.joinPath(path), DuraBasicTypesMapping.fromDuraName(child.getText()).type);
				
				break;

			case COMPOSITE_TYPE:
				try {
					attributes.put(DuraCQueryCompiler.joinPath(path),  typeDefinitions.getType(child.getText(), preamble));

					break;
				} catch (DuraUnspecificCompliationException e) {
					throw new DuraCompilationException(e, term);
				}

			default:
				throw new UnexpectedTokenException(child);
			}

		}
	}
	
	private static void addImplicitStreamAttributes(final Map<String, Type> attributes, final TemporalStreamSchemaTypeID streamType) {
		if (! attributes.containsKey(DuraDefaultAttributeNameMapping.identifierAttribute)) {
			attributes.put(DuraDefaultAttributeNameMapping.identifierAttribute, BasicType.instance(BasicTypeID.IDENTIFIER));
		}
				
		if (streamType==TemporalStreamSchemaTypeID.EVENT && !attributes.containsKey(eventIntervalName)) {
			attributes.put(eventIntervalName, TypeDefinitions.TIME_INTERVAL);
		}
	}

	private static FixedArrayNavigableMap<String, Type> getImplicitAttributes(final TemporalStreamSchemaTypeID streamType) {
		Map<String, Type> attributes = new HashMap<String, Type>();
		
		addImplicitStreamAttributes(attributes, streamType);
		
		return new FixedArrayNavigableMap<String, Type>(attributes);
	}
	
	
	private static void addDefaultConstraintsForStreamBounds(final ArrayList<ConditionFormula> relationConstraints, final TemporalStreamSchemaTypeID streamType) {
		IdentifierAttribute identifier = IdentifierAttribute.create(DuraDefaultAttributeNameMapping.identifierAttribute);
		ClockFunction id_seq_clock = new ClockFunction(DuraDefaultAttributeNameMapping.identifierClockName, BasicTypeID.IDENTIFIER);
		
		ConditionFormula id_constraint = new DefaultDefinition(identifier, id_seq_clock);
		
		if (! relationConstraints.contains(id_constraint)) {
			relationConstraints.add(id_constraint);
		}
		
		if (streamType == TemporalStreamSchemaTypeID.EVENT) {
			TimestampAttribute receptionTimeBegin = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(eventIntervalName, intervalBeginAttribute));
			TimestampAttribute receptionTimeEnd = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(eventIntervalName, intervalEndAttribute));
			ClockFunction system_time_clock = new ClockFunction(DuraDefaultAttributeNameMapping.sytemTimeClockName, BasicTypeID.TIMESTAMP);
			
			ConditionFormula receptionTimeBegin_constraint = new DefaultDefinition(receptionTimeBegin, system_time_clock);
			ConditionFormula receptionTimeEnd_constraint = new DefaultDefinition(receptionTimeEnd, system_time_clock);
			
			if (! relationConstraints.contains(receptionTimeBegin_constraint)) {
				relationConstraints.add(receptionTimeBegin_constraint);
			}
			
			if (! relationConstraints.contains(receptionTimeEnd_constraint)) {
				relationConstraints.add(receptionTimeEnd_constraint);
			}
		}
	}
	
	private static void addImplicitIdStreamBound(final ArrayList<ConditionFormula> relationConstraints, final TemporalStreamSchemaTypeID streamType) {
		IdentifierAttribute identifierAttribute = IdentifierAttribute.create(DuraDefaultAttributeNameMapping.identifierAttribute);
		ConditionFormula identifierStreamBound = new StreamBoundConstraint(identifierAttribute, true);
		
		if (!relationConstraints.contains(identifierStreamBound)) {
			relationConstraints.add(identifierStreamBound);
		}
	}

	private static void addImplicitReceptionTimeStreamBound(final ArrayList<ConditionFormula> relationConstraints, final TemporalStreamSchemaTypeID streamType) {
		if (streamType == TemporalStreamSchemaTypeID.EVENT) {
			TimestampAttribute receptionTimeEndAttribute = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(eventIntervalName, intervalEndAttribute));
			ConditionFormula receptionTimeEndStreamBound = new StreamBoundConstraint(receptionTimeEndAttribute);
			
			if (!relationConstraints.contains(receptionTimeEndStreamBound)) {
				relationConstraints.add(receptionTimeEndStreamBound);
			}
		}
	}
	
	private void addStatefulObjectInputStream(final String full_name, final DuraTree definition, final HashMap<String, Type> attributes, DuraVisibilityModifier visibility, final Collection<ConditionFormula> tupleConstraints, final Collection<ConditionFormula> relationConstraints) {
		try{
			StateStreamSchema streamSchema = new StateStreamSchema(new FixedArrayNavigableMap<String, Type>(attributes), new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
	
			TSA_TemporalStream stream = new TSA_TemporalStream(full_name, "", streamSchema, new FixedArrayNavigableSet<DTSA_BufferType>(DTSA_BufferType.INPUT), definition.getPositionDescription());
			
			if (this.stream_schema.get(TemporalStreamSchemaTypeID.STATE).containsKey(full_name)) {
				throw new DuraCompilationException(String.format("multiple definitions of stream %s found", full_name), definition);
			}
	
			this.stream_schema.get(TemporalStreamSchemaTypeID.STATE).put(full_name, new Definition<TSA_TemporalStream>(full_name, visibility, stream));
		} catch (TSA_SchemaException e) {
			throw new DuraCompilationException(e, definition);
		}
	}
	
	private static ArrayList<ConditionFormula> compileConstraints(final DuraTree constraints, final HashMap<String, Type> attributeTypes) {
		int conditionCount = constraints.getChildCount();
		ArrayList<ConditionFormula> conditions = new ArrayList<ConditionFormula>(conditionCount);
		
		Adapter<AtomicTerm<?>, AtomicTerm<? extends Type>> typeSchemaAttributes = new Adapter<AtomicTerm<?>, AtomicTerm<? extends Type>>() {
			
			@Override
			public AtomicTerm<? extends Type> adapt(final AtomicTerm<?> source) {
				if (source instanceof Attribute) {
					Attribute<?> attribute = (Attribute<?>) source;
					
					if (source.isTyped) {
						return (AtomicTerm<? extends Type>) attribute;
					} else {
						if (! attributeTypes.containsKey(attribute.name)) {
							throw new DuraCompilationException(String.format("unknown attribute: %s", attribute.name), constraints);
						}
						
						return TypedAttribute.create(attribute.name, attributeTypes.get(attribute.name));
					}
				} else {
					assert (source.isTyped);
					
					return (AtomicTerm<? extends Type>) source;
				}
			}
				
		};
		
		for (int i=0; i<conditionCount; i++) {
			conditions.add(DuraCQueryCompiler.compileFormula(constraints.getChild(i)).replaceAtomicTerm(typeSchemaAttributes));
		}
		
		return conditions;
	}
	
	private void addImpilictStreams(final DuraTree origin, final String full_name, final TemporalStreamSchemaTypeID streamType, DuraVisibilityModifier visibility, final TemporalStreamSchema explicitSchema, final String... suffixes) {
		this.addImpilictStreams(origin, full_name, streamType, visibility, explicitSchema, false, new FixedArrayNavigableSet<DTSA_BufferType>(DTSA_BufferType.WORKING), suffixes);
	}
	 
	private void addImpilictStreams(final DuraTree origin, final String full_name, final TemporalStreamSchemaTypeID streamType, DuraVisibilityModifier visibility, final TemporalStreamSchema explicitSchema, final boolean staticSchema, final FixedArrayNavigableSet<DTSA_BufferType> modifiers, final String... suffixes) {
		for (String suffix : suffixes) {
			try{
				String stream_name = String.format("%s%s%s", full_name, labelSeparator, suffix);
				
				ArrayList<ConditionFormula> tupleConstraints = new ArrayList<ConditionFormula>();
				ArrayList<ConditionFormula> relationConstraints = new ArrayList<ConditionFormula>();

				Definition<TSA_TemporalStream> streamDefinition;
				
				TemporalStreamSchema implicitSchema = explicitSchema.extend(getImplicitAttributes(streamType));
				
				tupleConstraints.addAll(implicitSchema.tupleConstraints().javaCollection());
				relationConstraints.addAll(implicitSchema.relationConstraints().javaCollection());
				
				addDefaultConstraintsForStreamBounds(relationConstraints, streamType);
				
				if (! staticSchema) {
					addImplicitIdStreamBound(relationConstraints, streamType);
					addImplicitReceptionTimeStreamBound(relationConstraints, streamType);
				}
				
				switch (streamType) {
				case EVENT:
					implicitSchema = new EventStreamSchema(implicitSchema.basicAttributes(), new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
					streamDefinition = new Definition<TSA_TemporalStream>(stream_name, visibility, new TSA_TemporalStream(stream_name, "", implicitSchema, modifiers, origin.getPositionDescription()));
					
					break;
					
				case ACTION:
					implicitSchema = new ActionStreamSchema(implicitSchema.basicAttributes(), new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
					streamDefinition = new Definition<TSA_TemporalStream>(stream_name, visibility, new TSA_TemporalStream(stream_name, "", implicitSchema, modifiers, origin.getPositionDescription()));
	
					this.addImpilictActionStreams(origin, streamDefinition);
					
					return;			//FIXME: correct?
	
				default:
					throw new IllegalArgumentException();
				}
				
				this.stream_schema.get(streamType).put(stream_name, streamDefinition);
			} catch (TSA_SchemaException e) {
				throw new DuraCompilationException(e, origin);
			}
		}
	}
	
	private void addImpilictActionStreams(final DuraTree origin, Definition<TSA_TemporalStream> action) {		
		try{
			FixedArrayNavigableMap<String, String> payloadPrefix = new FixedArrayNavigableMap<String, String>(new String[]{ "" }, new String[] { payloadLabel });
			
			TemporalStreamSchema initiatedSchema = action.definition.schema.rename(payloadPrefix);
			TemporalStreamSchema commonSchema = new TemporalStreamSchema(initiatedSchema.attributes, new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());

			this.addImpilictStreams(origin, action.name, TemporalStreamSchemaTypeID.EVENT, DuraVisibilityModifier.DEFAULT, initiatedSchema, actionInitiatedSuffix);
			this.addImpilictStreams(origin, action.name, TemporalStreamSchemaTypeID.EVENT, DuraVisibilityModifier.DEFAULT, commonSchema, actionInitiatedSuffix, actionSucceededSuffix, actionFailedSuffix);
		} catch (TSA_SchemaException e) {
			throw new DuraCompilationException(e, origin);
		}
	}
	
	private void addImpilictStatefulObjectStreams(final DuraTree origin, Definition<TSA_TemporalStream> statefulObject, final FixedArrayNavigableSet<DTSA_BufferType> modifiers) {
		try{
			TemporalStreamSchema statefulObjectSchema = statefulObject.definition.schema;
			
			FixedArrayNavigableMap<String, String> payloadPrefix = new FixedArrayNavigableMap<String, String>(new String[]{ "" }, new String[] { payloadLabel });
			FixedArrayNavigableMap<String, String> obsoletePayloadPrefix = new FixedArrayNavigableMap<String, String>(new String[]{ "" }, new String[] { obsoletePayloadLabel });
			FixedArrayNavigableMap<String, String> valuesPrefix = new FixedArrayNavigableMap<String, String>(new String[]{ "" }, new String[] { valuesLabel });
			
			//entailed events
			FixedArrayNavigableMap<String, Type> provenanceIdAttribute = new FixedArrayNavigableMap<String, Type>( 
					new String[]{ TSA_SyntaxPatterns.attributeName(provenanceLabel, identifierAttribute) }, 
					new Type[] { BasicTypeID.IDENTIFIER.type } 
				);
	
			TemporalStreamSchema basicEventSchema = statefulObjectSchema.rename(payloadPrefix).extend(provenanceIdAttribute);
			TemporalStreamSchema terminatedSchema = new TemporalStreamSchema(basicEventSchema.attributes, new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
			TemporalStreamSchema updatedSchema = new TemporalStreamSchema(FixedArrayNavigableMap.merge(basicEventSchema.attributes, statefulObjectSchema.rename(obsoletePayloadPrefix).attributes), new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
			
			this.addImpilictStreams(origin, statefulObject.name, TemporalStreamSchemaTypeID.EVENT, statefulObject.visibility, basicEventSchema, statefulObjectSchema.isStatic, modifiers, soCreatedSuffix);
			this.addImpilictStreams(origin, statefulObject.name, TemporalStreamSchemaTypeID.EVENT, statefulObject.visibility, terminatedSchema, statefulObjectSchema.isStatic, modifiers, soTerminatedSuffix);
			this.addImpilictStreams(origin, statefulObject.name, TemporalStreamSchemaTypeID.EVENT, statefulObject.visibility, updatedSchema, statefulObjectSchema.isStatic, modifiers, soUpdatedSuffix);
	
			if (! statefulObjectSchema.isStatic) {
				//implicit actions
				FixedArrayNavigableMap<String, Type> valuesAttributes = statefulObjectSchema.retain(FixedArrayNavigableSet.difference(statefulObjectSchema.attributes.keys(), getImplicitAttributes(TemporalStreamSchemaTypeID.STATE).keys())).rename(valuesPrefix).attributes;
		
				FixedArrayNavigableMap<String, Type> queryAttributes = new FixedArrayNavigableMap<String, Type>(
						new String[]{ TSA_SyntaxPatterns.attributeName(queryLabel, identifierAttribute) }, 
						new Type[] { BasicTypeID.IDENTIFIER.type } 
					);
		
				TemporalStreamSchema createSchema = new TemporalStreamSchema(valuesAttributes, new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
				TemporalStreamSchema terminateSchema = new TemporalStreamSchema(queryAttributes, new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
				TemporalStreamSchema updateSchema = new TemporalStreamSchema(FixedArrayNavigableMap.merge(valuesAttributes, queryAttributes), new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
		
				this.addImpilictStreams(origin, statefulObject.name, TemporalStreamSchemaTypeID.ACTION, DuraVisibilityModifier.DEFAULT, createSchema, soCreateSuffix);
				this.addImpilictStreams(origin, statefulObject.name, TemporalStreamSchemaTypeID.ACTION, DuraVisibilityModifier.DEFAULT, terminateSchema, soTerminateSuffix);
				this.addImpilictStreams(origin, statefulObject.name, TemporalStreamSchemaTypeID.ACTION, DuraVisibilityModifier.DEFAULT, updateSchema, soUpdateSuffix);
			}
		} catch (TSA_SchemaException e) {
			throw new DuraCompilationException(e, origin);
		}
	}
}
