package de.lmu.ifi.pms.cep.dura.compiler.durac.collections;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.BASIC_TYPE;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.COMPOSITE_TYPE;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.PREAMBLE;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.TERM;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.TYPE;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.TYPEDEF;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalBeginAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalEndAttribute;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.TreeIterator;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCQueryCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraBasicTypesMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraVisibilityModifier;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnknownTypeException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationRuntimeException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraVisibilityException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


/**
 * Contains a mapping from names of user defined dura types to TSA types.
 *   
 * @author hausmann
 *
 */
public class TypeDefinitions extends HashMap<String, Definition<Type>> {

	public static final CompositeType TIME_POINT;
	public static final NamedType<CompositeType> TIME_INTERVAL;
	
	static {
		try {
			HashMap<String, Type> attributes = new HashMap<String, Type>();
			
			attributes.put(intervalBeginAttribute, BasicType.instance(BasicTypeID.TIMESTAMP));
			
			TIME_POINT = CompositeType.get(attributes);
			
			attributes.put(intervalEndAttribute, BasicType.instance(BasicTypeID.TIMESTAMP));
			
			FixedArray<ConditionFormula> intervalConstraints = new FixedArray<ConditionFormula>(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, TimestampAttribute.create(intervalBeginAttribute), TimestampAttribute.create(intervalEndAttribute)));
			
			TIME_INTERVAL = NamedType.get("time-interval", CompositeType.get(attributes, intervalConstraints));
		
		} catch (TSA_TypeException e) {
			throw new DuraUnspecificCompliationRuntimeException(e);
		}
	}
	
	private TypeDefinitions() { }
	
	public TypeDefinitions(final DuraTree prmbl, final Preamble preamble) {
		assert(prmbl.getType() == PREAMBLE);
		
		for (DuraTree definition : new TreeIterator(prmbl, TYPEDEF)) {
			String name = TSA_SyntaxPatterns.bufferFullName(preamble.package_name, definition.getText());
			DuraVisibilityModifier visibility = DuraVisibilityModifier.toVisibilityModifier(definition.getChild(0));
			Type type = this.parseTypeTerm(definition.getChild(1), preamble);
			
			this.put(name, new Definition<Type>(name, visibility, type));
		}
	}

	public Type parseTypeTerm(final DuraTree term, final Preamble preamble) {
		assert(term.getType() == TERM ||
				term.getType() == TYPE);

		switch (term.getChild(0).getType()) {
		case BASIC_TYPE:
			return DuraBasicTypesMapping.fromDuraName(term.getChild(0).getText()).type;

		case COMPOSITE_TYPE:
			try {
				return this.getType(term.getChild(0).getText(), preamble);
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCompilationException(e, term);
			}

		case TERM:
			int childCount = term.getChildCount();
			DuraTree tupleConstraints = term.getChild(childCount-1).getChild(0);
			
			HashMap<String, Type> attributes = new HashMap<String, Type>();
			
			for (int i=0; i<childCount-1; i++) {
				DuraTree child = term.getChild(i);
				
				attributes.put(child.getText(), this.parseTypeTerm(child, preamble));
			}
			
			try {
				if (tupleConstraints != null) {
					return CompositeType.get(attributes, compileConstraints(tupleConstraints));
				} else {
					return CompositeType.get(attributes);
				}
			} catch (TSA_TypeException e) {
				throw new DuraCompilationException(e, term);
			}
			
		default:
			throw new UnexpectedTokenException(term.getChild(0));
		}
	}
	
	private static FixedArray<ConditionFormula> compileConstraints(final DuraTree constraints) {
		int conditionCount = constraints.getChildCount();
		ConditionFormula[] conditions = new ConditionFormula[conditionCount];
		
		for (int i=0; i<conditionCount; i++) {
			conditions[i] = DuraCQueryCompiler.compileFormula(constraints.getChild(i));
		}
		
		return new FixedArray<ConditionFormula>(conditions);
	}
	
	public Type getType(final String typeName, final Preamble preamble) throws DuraUnknownTypeException, DuraVisibilityException {
		Definition<Type> definition = getDefinition(typeName, preamble);
		
		if (definition.visibility==DuraVisibilityModifier.PUBLIC || definition.name.startsWith(preamble.package_name)) {
			return definition.definition;
		} else {
			throw new DuraVisibilityException(definition.name);
		}
	}
	
	private Definition<Type> getDefinition(String typeName, Preamble preamble) throws DuraUnknownTypeException {
		if (typeName.equals("time-interval")) {
			return new Definition<Type>("time-interval", DuraVisibilityModifier.PUBLIC, TIME_INTERVAL);
		} else if (typeName.contains(".")) {
			if (! this.containsKey(typeName)) {
				throw new DuraUnknownTypeException(typeName);
			}
			
			return this.get(typeName);
		} else if (this.containsKey(TSA_SyntaxPatterns.bufferFullName(preamble.package_name, typeName))){
			String full_name = TSA_SyntaxPatterns.bufferFullName(preamble.package_name, typeName);
			
			return this.get(full_name);
		} else if (preamble.imports.containsKey(typeName)) {
			String full_name = preamble.imports.get(typeName);
			
			return this.get(full_name);
		}
		
		throw new DuraUnknownTypeException(typeName);
	}
	
	public static TypeDefinitions join(final List<TypeDefinitions> definitions) {
		TypeDefinitions result = new TypeDefinitions();
		
		for (TypeDefinitions definition : definitions) {
			assert(Collections.disjoint(result.keySet(), definition.keySet()));
			
			result.putAll(definition);
		}
		
		return result;
	}
}