package de.lmu.ifi.pms.cep.event_mill.configuration;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.Path;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.strategy.ExplicitCycleStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.io.bus.ExplicitMessageBusProviderCollection;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBusProvider;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBusProviderCollection;
import de.lmu.ifi.pms.cep.event_mill.io.bus.WrapperMessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalog;
import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalogProvider;
import de.lmu.ifi.pms.cep.event_mill.io.catalog.EmptyDataCatalogProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.ExplicitMessageFormatProviderCollection;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.ExplicitStreamIOProviderCollection;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIOProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIOProviderCollection;

import de.lmu.ifi.pms.java.util.collection.Collection;

@Root(name="configuration")
public class IOConfiguration {
	
	@Path(value="io")
	@Element(name="streams", required=false)
	StreamIOProviderCollection streams;
	
	@Path(value="io")
	@Element(name="formats", required=false)
	final ExplicitMessageFormatProviderCollection formats;
	
	@Path(value="io")
	@Element(name="buses", required=false)
	final MessageBusProviderCollection buses;

	@Path(value="io")
	@Element(name="data-catalogs", required=false)
	final DataCatalogProvider dataCatalogs;
	
	public IOConfiguration(final Collection<StreamIOProvider> streams, final Collection<MessageFormatProvider<?>> formats, final Collection<MessageBusProvider> buses) {
		this((streams == null ? new ExplicitStreamIOProviderCollection(streams) : new ExplicitStreamIOProviderCollection(streams)), (formats == null ? new ExplicitMessageFormatProviderCollection() : new ExplicitMessageFormatProviderCollection(formats)), (buses == null? new ExplicitMessageBusProviderCollection() : new ExplicitMessageBusProviderCollection(buses)), null);		
	}
	
	public IOConfiguration(
			@Path(value="io")
			@Element(name="streams", required=false) final
			StreamIOProviderCollection streams,
			@Path(value="io")
			@Element(name="formats", required=false)
			final ExplicitMessageFormatProviderCollection formats,
			@Path(value="io")
			@Element(name="buses", required=false) final
			MessageBusProviderCollection buses,
			@Path(value="io")
			@Element(name="data-catalogs", required=false)
			final DataCatalogProvider dataCatalogs
			) {
		super();
		
		if(streams == null){
			this.streams = new ExplicitStreamIOProviderCollection();
		}
		else{
			this.streams = streams;
		}
		if(formats == null){
			this.formats = new ExplicitMessageFormatProviderCollection();
		}
		else{
			this.formats = formats;
		}
		if(buses == null){
			this.buses = new ExplicitMessageBusProviderCollection();
		}
		else{
			this.buses = buses;
		}
		if(dataCatalogs == null){
			this.dataCatalogs = new EmptyDataCatalogProvider();
		}
		else{
			this.dataCatalogs = dataCatalogs;
		}
	}
	
	public StreamIOProvider streamProvider(final String streamProviderIdentifier){
		StreamIOProvider streamIOProvider = this.streams.get(streamProviderIdentifier);
		if(streamIOProvider != null){
			return streamIOProvider;
		}
		else{
			throw new RuntimeException("Stream IO provider \"" + streamProviderIdentifier + "\" does not exist.");
		}
	}
	
	public StreamIO stream(final String streamProviderIdentifier, final String identity){
		return this.streamProvider(streamProviderIdentifier).connect(identity);
	}
	
	public MessageBusProvider bus(final String busIdentifier){
		MessageBusProvider messageBusProvider = this.buses.get(busIdentifier);
		if(messageBusProvider != null){
			return messageBusProvider;
		}
		else{
			throw new RuntimeException("Message bus provider \"" + busIdentifier + "\" does not exist.");
		}
	}
	
	public MessageFormatProvider<?> format(final String formatIdentifier){
		MessageFormatProvider<?> messageFormatProvider = this.formats.get(formatIdentifier);
		if(messageFormatProvider != null){
			return messageFormatProvider;
		}
		else{
			throw new RuntimeException("Message format \"" + formatIdentifier + "\" does not exist.");
		}
	}
	
	@SuppressWarnings("unchecked")
	public <MessageType> MessageFormatProvider<MessageType> format(final Class<MessageType> messageType, final String formatIdentifier){
		MessageFormatProvider<?> format = this.format(formatIdentifier);
		if(messageType.isAssignableFrom(format.type)){
			return (MessageFormatProvider<MessageType>) format; 
		}
		else{
			throw new RuntimeException("Type of format \"" + formatIdentifier + "\" is incompatible with \"" + messageType.getName() + "\"");
		}
	}
	
	public DataCatalog dataCatalog(final String catalogIdentifier){
		DataCatalog dataCatalog = this.dataCatalogs.create(catalogIdentifier);
		if(dataCatalog != null){
			return dataCatalog;
		}
		else{
			throw new RuntimeException("Data catalog \"" + catalogIdentifier + "\" does not exist.");
		}
	}
	
	public <MessageType> MessageBus<MessageType> bus(final String busIdentifier, final Class<MessageType> messageType, final String formatIdentifier, final String componentIdentifer){
		final MessageFormatProvider<MessageType> format = this.format(messageType, formatIdentifier);
		return this.bus(busIdentifier, messageType, format, componentIdentifer);
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public <MessageType> MessageBus<MessageType> bus(final String busIdentifier, final Class<MessageType> messageType, final MessageFormatProvider<MessageType> format, final String componentIdentifer){
		MessageBusProvider messageBusProvider = this.bus(busIdentifier);
		MessageBus<?> messageBus = messageBusProvider.connect(componentIdentifer);
		if(messageBus.formatter().identifier.equals(format.identifier)){
			assert messageType.isAssignableFrom(messageBus.formatter().type);
			assert messageType.isAssignableFrom(messageBus.type);
			return (MessageBus<MessageType>) messageBus;
		}
		else{
			DataCatalog dataCatalog = this.dataCatalog(busIdentifier);
			return new WrapperMessageBus(format, dataCatalog, messageBus);
		}
	}

	public static final Logger logger = LoggerFactory.getLogger(IOConfiguration.class);

	public static boolean exists(){
		return existsResource(DEFAULTS.defaultIOConfigResourcePath);
	}
	
	public static IOConfiguration read() throws IOException{
		return readResource(DEFAULTS.defaultIOConfigResourcePath);
	}
	
	public static String defaultConfigFilePath(){
		return DEFAULTS.resourceFilePath(DEFAULTS.defaultIOConfigResourcePath);
	}
	
	public static void write(final IOConfiguration configuration){
		write(defaultConfigFilePath(), configuration);
	}
	
	public static boolean existsFile(final String path){
		File configFile = new File(path);
		return configFile.exists();
	}
	
	public static boolean existsResource(final String path){		
		return IOConfiguration.class.getClassLoader().getResource(path) != null;
	}
	
	public static IOConfiguration readResource(final String path) throws IOException{
		InputStream is = IOConfiguration.class.getClassLoader().getResourceAsStream(path);
		if(is != null){
			return read(is);
		}
		else{
			throw new IOException("Could not find resource \"" + path + "\".");
		}
	}
	
	public static IOConfiguration readFile(final String path) throws IOException{
		return read(new FileInputStream(path));
	}
	
	public static IOConfiguration read(final InputStream input) throws IOException{
		Serializer serializer = new Persister(new ExplicitCycleStrategy("iri","ref"), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		try {
			return serializer.read(IOConfiguration.class, input);
		} catch (RuntimeException e) {
			throw e;
		} catch (IOException e) {
			throw e;
		} catch (Exception e) {
			logger.error("Unable to read configuration file!", e);
			throw new RuntimeException(e);
		}
	}
	
	public static void write(final String path, final IOConfiguration configuration){
		Serializer serializer = new Persister(new ExplicitCycleStrategy("iri","ref"), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		File configFile = new File(path);
		try {
			if(!configFile.exists()){
				configFile.createNewFile();
			}
			serializer.write(configuration, configFile);
		} catch (Exception e) {
			logger.error("Unable to write configuration file!", e);
			throw new RuntimeException(e);
		}
	}
	
	public static void main(final String[] args) throws Exception {
		System.out.println(defaultConfigFilePath());
		IOConfiguration io = IOConfiguration.read();
		Serializer s = new Persister(new ExplicitCycleStrategy("iri","ref"), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		s.write(io, System.out);
	}
	
}
