package de.lmu.ifi.pms.cep.event_mill.configuration;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Statement;

import org.simpleframework.xml.Root;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.development.UnimplementedMethodError;

import de.lmu.ifi.pms.java.LoggingObject;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;

@Root(name="driver")
public abstract class SQL_DatabaseDriver extends LoggingObject{
	
	public final String jdbcDriver;
	public final String subProtocol;
	public final Logger internalLogger;
	
	public static final FixedArray<String> log4jdbcLoggers = new FixedArray<String>("jdbc.sqlonly", "jdbc.sqltiming", "jdbc.audit", "jdbc.resultset", "jdbc.connection");
	
	public static enum LogLevel{
		TRACE,
		DEBUG,
		INFO,
		WARN,
		ERROR,
		OFF
		;
		
		public static LogLevel level(final Collection<String> loggers){
			LogLevel minLevel = OFF;
			for(String logger: loggers){
				LogLevel level = level(logger);
				if(level.ordinal() < minLevel.ordinal()){
					minLevel = level;
				}				
			}
			return minLevel;
		}
		
		public static LogLevel level(final String... loggers){
			LogLevel minLevel = OFF;
			for(String logger: loggers){
				LogLevel level = level(logger);
				if(level.ordinal() < minLevel.ordinal()){
					minLevel = level;
				}				
			}
			return minLevel;
		}
		
		public static LogLevel level(final String logger){
			return level(LoggerFactory.getLogger(logger));
		}
		
		public static LogLevel level(final Logger logger){
			if(logger.isTraceEnabled()){
				return TRACE;
			}
			else if(logger.isDebugEnabled()){
				return DEBUG;
			}
			else if(logger.isInfoEnabled()){
				return INFO;
			}
			else if(logger.isWarnEnabled()){
				return WARN;
			}
			else if(logger.isErrorEnabled()){
				return ERROR;
			}
			else{
				return OFF;
			}
		}
	}
	
	@Override
	protected Logger logger() {
		throw new UnimplementedMethodError();
		//return null;//FIXME implement
	}
	
	public SQL_DatabaseDriver(final String jdbcDriver, final String subProtocol) throws ClassNotFoundException{
		super();
		this.jdbcDriver = jdbcDriver;
		registerDriver(jdbcDriver);
		this.subProtocol = subProtocol;
		Class<?> driverClass = Class.forName(jdbcDriver, true, LiberiesLoader.classLoader);
		try {
			//DriverManager.registerDriver((Driver)driverClass.getConstructor().newInstance());
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		this.internalLogger = LoggerFactory.getLogger(driverClass);
	}	
	
	public LogLevel internalLogLevel(){
		return LogLevel.level(this.internalLogger);
	}
	
	public LogLevel externalLogLevel(){
		return LogLevel.level(log4jdbcLoggers);
	}
	
	public String url(final String serverUrl, final String database){
		StringBuilder sb = new StringBuilder();
		sb.append("jdbc:");
		if(this.externalLogLevel() != LogLevel.OFF & log4jdbcIsAvailable){
			sb.append("log4jdbc:");
		}
		sb.append(this.subProtocol);
		sb.append("://");
		sb.append(serverUrl);
		if(database != null){
			sb.append(this.urlDatabaseExpansion(database));
		}
		return sb.toString();
	}
	
	protected abstract String urlDatabaseExpansion(final String database);
	
	protected abstract boolean isSetLogging(java.util.Properties properties);
	
	protected String logDestination(){
		return DEFAULTS.logFileDir + DEFAULTS.logTime + "-" + this.subProtocol + ".log";
	}
	
	protected abstract void setLoggingProperty(java.util.Properties properties, LogLevel logLevel, String logDestination);

	public Connection connect(final SQL_ConnectionConfiguration configuration) throws SQLException{
		java.util.Properties properties = configuration.properties();
		if(!this.isSetLogging(properties)){
			this.setLoggingProperty(properties, this.internalLogLevel(), this.logDestination());
		}
		
		Connection connection = this.connect(this.url(configuration.serverUrl, configuration.database), properties);
		
		try{
			connection.setAutoCommit(configuration.autoCommitMode);
		} catch(SQLException e){
			this.error("Error on setting auto commit mode to true!", e);
			throw new RuntimeException(e);
		}
//		try{
//			this.connection.setTransactionIsolation(Connection.TRANSACTION_NONE);
//		} catch(SQLException e){
//			this.error("Error on setting transaction isolation!", e);
//			throw new RuntimeException(e);
//		}
		try{
			if(connection.getMetaData().supportsCatalogsInTableDefinitions() && configuration.catalog != null && !configuration.catalog.equals("")){
				connection.setCatalog(configuration.catalog);
			}
		} catch(SQLException e){
			this.error("Error on setting catalog!", e);
			throw new RuntimeException(e);
		}
		
		return connection;
	}
	
	protected Connection connect(final String jdbcUrl, final java.util.Properties properties) throws SQLException{
		return DriverManager.getConnection(jdbcUrl, properties);
	}
	
	public String nameOfGreatestFunction(){
		return "GREATEST";
	}
	
	public void initGreatestFunctions(final Statement statement) throws SQLException{
		
	}
	
	public void clearGreatestFunctions(final Statement statement) throws SQLException{
		
	}
	
	public String nameOfLeastFunction(){
		return "LEAST";
	}
	
	public void initLeastFunctions(final Statement statement) throws SQLException{
		
	}
	
	public void clearLeastFunctions(final Statement statement) throws SQLException{
		
	}
		
	protected static final Logger logger = LoggerFactory.getLogger(SQL_DatabaseDriver.class);
	
	protected static final boolean log4jdbcIsAvailable;
	
	static{
		boolean success;
		try{
			//LiberiesLoader.classLoader.loadClass("de.lmu.ifi.pms.cep.Version");
			registerDriver("net.sf.log4jdbc.DriverSpy");
			success = true;
		}
		catch (Exception e) {
			logger.error("Driver net.sf.log4jdbc.DriverSpy could not be loaded!", e);
			success = false;
		}
		log4jdbcIsAvailable = success;
	}
	
	public static synchronized void registerDriver(final String driver){
		try {
			Class<?> c = Class.forName(driver, true, LiberiesLoader.classLoader);
			try {
				java.util.Collection<Driver> usualDrivers = java.util.Collections.list(DriverManager.getDrivers());
				for(Driver regDriver: java.util.Collections.list(LiberiesLoader.classLoader.driverReader.getDrivers())){
					if(!(regDriver instanceof DriverWrapper)){
						if(!usualDrivers.contains(regDriver)){
							try {
								DriverManager.registerDriver(new DriverWrapper(regDriver));
								LiberiesLoader.classLoader.driverReader.deregisterDriver(regDriver);
							} catch (SQLException e) {
								logger.warn("",e);
							}
						}
					}
				}
			} catch (Exception e) {
				logger.warn("",e);
			}
			
			logger.debug("Loaded driver {}", c.getName());
		} catch (ClassNotFoundException e) {
			logger.error("Driver " + driver + "could not be loaded!", e);
			throw new IllegalArgumentException("Driver not available!",e);
		}
	}
	
	public static synchronized boolean hasDriver(final String url){
		try {
			checkDriver(url);
			return true;
		} catch (IllegalArgumentException e) {
			return false;
		}
	}
	
	public static synchronized void checkDriver(final String url){
		try {
			DriverManager.getDriver(url);
		} catch (SQLException e) {
			logger.error("No driver available for the url!", e);
			throw new IllegalArgumentException("No driver for that url \"" + url + "\" available!",e);
		}
	}

}
