package de.lmu.ifi.pms.cep.event_mill.io.format;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class StringPatternGenerator extends MessageGenerator<String>{
	
	final static String defaultEscapeSequence = "\\";
	final static String defaultParameterStartSequence = "%{";
	final static String defaultParameterEndSequence = "}";
	final static ParameterFormatter defaultFormater = new ParameterFormatter();

	public final String pattern;
	final String[] parameterValues;
	final int[] parameterReferences;
	final String[] splitedPattern;
	final java.lang.StringBuilder sb;
	
	public final String escapeSequence;
	public final String escapeRegEx;
	public final String parameterStartSequence;
	public final String parameterStartRegEx;
	public final String parameterEndSequence;
	public final String parameterEndRegEx;
	
	public final String parameterRegEx;
	public final Pattern parameterPattern;
	
	public final String unquoteStringRegEx;
	public final Pattern unquoteStringPattern;
	
	public final String unquoteParameterRegEx;
	public final Pattern unquoteParameterPattern;
	
	public final ParameterFormatter formatter;
	
	public String quoteString(final String s){
		return this.unquoteStringPattern.matcher(s).replaceAll(Matcher.quoteReplacement(this.escapeSequence) + "$1");
	}
	
	public String quoteParameter(final String parameter){
		return this.unquoteParameterPattern.matcher(parameter).replaceAll("$1");
	}
	
	public String unquoteString(final String s){
		return this.unquoteStringPattern.matcher(s).replaceAll("$1");
	}
	
	public String unquoteParameter(final String parameter){
		return this.unquoteParameterPattern.matcher(parameter).replaceAll("$1");
	}
	
	public StringPatternGenerator(final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(defaultFormater, pattern, parameters, types);
	}
	
	public StringPatternGenerator(final ParameterFormatter formater, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(formater, pattern, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types));
	}
	
	public StringPatternGenerator(final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(defaultFormater, pattern, types);
	}
	
	public StringPatternGenerator(final ParameterFormatter formater, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(formater, pattern, types.keys(), types.values());
	}
	
	public StringPatternGenerator(final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(defaultFormater, pattern, parameters, types);
	}
	
	public StringPatternGenerator(final ParameterFormatter formater, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(defaultEscapeSequence, defaultParameterStartSequence, defaultParameterEndSequence, formater, pattern, parameters, types);
	}
	
	public StringPatternGenerator(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, parameters, types);
	}
	
	public StringPatternGenerator(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, formater, pattern, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types));
	}
	
	public StringPatternGenerator(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, types);
	}
	
	public StringPatternGenerator(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, formater, pattern, types.keys(), types.values());
	}
	
	public StringPatternGenerator(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, parameters, types);
	}

	public StringPatternGenerator(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final ParameterFormatter formatter, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		super(String.class, parameters, types);
		this.escapeSequence = escapeSequence;
		this.escapeRegEx = Pattern.quote(this.escapeSequence);
		this.parameterStartSequence = parameterStartSequence;
		this.parameterStartRegEx = Pattern.quote(this.parameterStartSequence);
		this.parameterEndSequence = parameterEndSequence;
		this.parameterEndRegEx = Pattern.quote(this.parameterEndSequence);
		
		this.parameterRegEx = "(?<!" + this.escapeRegEx + ")" + this.parameterStartRegEx + "(.*?)" + "(?<!" + this.escapeRegEx + ")" + this.parameterEndRegEx;

		this.parameterPattern = Pattern.compile(this.parameterRegEx);
		this.unquoteStringRegEx = this.escapeRegEx + "(" + this.escapeRegEx + "|" + this.parameterStartRegEx + ")";
		this.unquoteStringPattern = Pattern.compile(this.unquoteStringRegEx);
		this.unquoteParameterRegEx = this.escapeRegEx + "(" + this.escapeRegEx + "|" + this.parameterEndRegEx + ")";
		this.unquoteParameterPattern = Pattern.compile(this.unquoteParameterRegEx);
		
		assert formatter != null;
		this.formatter = formatter;
		
		this.pattern = pattern;
		
		this.parameterValues = new String[this.parameters.size()];
		
		
		Matcher parameterMatcher = this.parameterPattern.matcher(pattern);
		ArrayList<String> splitedPattern = new ArrayList<String>();
		ArrayList<Integer> parameterReferences = new ArrayList<Integer>();
		int startIndex = 0;
		String parameter;
		int parameterIndex;
		if(parameterMatcher.find()){
			assert parameterMatcher.end() >= startIndex;
			splitedPattern.add(this.unquoteString(pattern.substring(startIndex, parameterMatcher.start())));
			startIndex = parameterMatcher.end();
			parameter = this.unquoteParameter(parameterMatcher.group(1));
			parameterIndex = this.indexOf(parameter);
			if(parameterIndex < 0){
				throw new IllegalArgumentException("Reference to undeclared parameter \"" + parameter + "\"");
			}
			assert parameterIndex >= 1;
			parameterReferences.add(parameterIndex - 1);
						
			while(parameterMatcher.find()){
				assert parameterMatcher.end() >= startIndex;
				splitedPattern.add(this.unquoteString(pattern.substring(startIndex, parameterMatcher.start())));
				startIndex = parameterMatcher.end();
				parameter = this.unquoteParameter(parameterMatcher.group(1));
				parameterIndex = this.indexOf(parameter);
				if(parameterIndex < 0){
					throw new IllegalArgumentException("Reference to undeclared parameter \"" + parameter + "\"");
				}
				assert parameterIndex >= 1;
				parameterReferences.add(parameterIndex - 1);
			}
			splitedPattern.add(this.unquoteString(pattern.substring(startIndex, pattern.length())));
			this.splitedPattern = splitedPattern.toArray(new String[splitedPattern.size()]);
		}
		else{
			this.splitedPattern = new String[]{pattern};
		}		
		
		//this.parameterReferences = parameterReferences.toArray(new int[parameterReferences.size()]);
		this.parameterReferences = new int[parameterReferences.size()];
		for(int parameterRef = 0; parameterRef < this.parameterReferences.length; ++parameterRef){
			this.parameterReferences[parameterRef] = parameterReferences.getInt(parameterRef);
		}
		
		assert this.splitedPattern.length == this.parameterReferences.length + 1;
		this.sb = new java.lang.StringBuilder();
	}
	
	public void compose(final StringBuilder sb, final String[] parameterValues){
		assert parameterValues.length == this.parameterValues.length;
		for(int parameterRef = 0; parameterRef < this.parameterReferences.length; ++parameterRef){
			sb.append(this.splitedPattern[parameterRef]);
			String value = parameterValues[this.parameterReferences[parameterRef]];
			if(value == null){
				sb.append("NULL");
			}
			else{
				sb.append(value);
			}
		}
		sb.append(this.splitedPattern[this.parameterReferences.length]);
	}
		
	public String compose(final String[] parameterValues){
		assert parameterValues.length == this.parameterValues.length;
		StringBuilder sb = new StringBuilder();
		this.compose(sb, parameterValues);
		return sb.toString();
	}	
	
	@Override
	public String compose(){
		if(this.sb.length() == 0){
			this.compose(this.sb, this.parameterValues);
		}
		return this.sb.toString();
	}	
	
	@Override
	public String toString() {
		return this.compose();
	}
	
	@Override
	public void clearParameters() {
		Arrays.fill(this.parameterValues, null);
		this.sb.setLength(0);
	}

	
	@Override
	public void setNull(final int parameterIndex, final int sqlType) {
		this.parameterValues[parameterIndex - 1] = null;
		this.sb.setLength(0);
	}

	
	@Override
	public void setBoolean(final int parameterIndex, final boolean x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatBoolean(x);
		this.sb.setLength(0);
	}

	
	@Override
	public void setByte(final int parameterIndex, final byte x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatByte(x);
		this.sb.setLength(0);
	}

	
	@Override
	public void setShort(final int parameterIndex, final short x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatShort(x);
		this.sb.setLength(0);
	}

	
	@Override
	public void setInt(final int parameterIndex, final int x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatInt(x);
		this.sb.setLength(0);
	}

	
	@Override
	public void setLong(final int parameterIndex, final long x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatLong(x);
		this.sb.setLength(0);
	}
	
	@Override
	public void setTimestamp(final int parameterIndex, final long x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatTimestamp(x);
		this.sb.setLength(0);
	}
	
	@Override
	public void setDuration(final int parameterIndex, final long x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatDuration(x);
		this.sb.setLength(0);
	}
	
	@Override
	public void setIdentifier(final int parameterIndex, final long x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatIdentifier(x);
		this.sb.setLength(0);
	}

	
	@Override
	public void setFloat(final int parameterIndex, final float x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatFloat(x);
		this.sb.setLength(0);
	}

	
	@Override
	public void setDouble(final int parameterIndex, final double x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatDouble(x);
		this.sb.setLength(0);
	}
	
	@Override
	public void setString(final int parameterIndex, final String x) {
		this.parameterValues[parameterIndex - 1] = this.formatter.formatString(x);
		this.sb.setLength(0);
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(StringPatternGenerator.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

	public static void main(final String[] args) {
		String pattern = "person{ name{%{person.%{name}}, address{ street{%{person.\\}address.street}}, number{%{person.a\\\\ddress.number}}}} }";
		
		StringPatternGenerator generator = new StringPatternGenerator(pattern, 
				new String[]{"person.%{name", "person.}address.street", "person.a\\ddress.number"},
				new BasicTypeID[]{	STRING, STRING, INT}
				);
		
		generator.setString(1, "nobody");
		generator.setString(2, "Valeria");
		generator.setInt(3, 15);
		
		System.out.println(generator.toString());
		
		generator.clearParameters();
	}

}
