package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.time.Clock;


public class BufferBlockCounterClock extends Clock {
	
	protected static final Logger logger = LoggerFactory.getLogger(BufferBlockCounterClock.class); 
	
	protected final SQL_StreamBufferStore bufferStore;
	protected final String getValueSQL;
	protected final String incrementValueSQL;
	protected PreparedStatement getValue;
	protected PreparedStatement incrementValue;
	boolean usePreparedStatements;
	
	public BufferBlockCounterClock(final SQL_StreamBufferStore bufferStore, final String program_meta_schema, final DTSA_BufferType type, final String buffer_full_name, final boolean usePreparedStatements){
		assert program_meta_schema.matches(SQL_Translation.schema_name_RegEx);
		assert buffer_full_name.matches(TSA_SyntaxPatterns.BufferFullNameRegEx);
		
		this.bufferStore = bufferStore;
		int buffer_id;
		
		
		Connection c = null;
		
		try{
			while(true){
				try{
						ResultSet id;
						c = bufferStore.connection();
						id = c.createStatement().executeQuery(SQL_Translation.sqlGetBufferId(program_meta_schema, type, buffer_full_name));
						id.next();
						buffer_id = SQL_Translation.sqlReadBufferMetaEntryId(id);
						break;
				}
				catch(SQLException e){
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						logger.debug("Connection failure. Trying again.", e);
						assert c != null;
						this.bufferStore.reconnect(c);
						this.init();
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						logger.debug("Reading block counter rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						logger.debug("Handle of read block counter statement is invalid. Repreparing statement and trying again.", e);
						this.init();
					}
					else{
						throw e;
					}
				}				
			}
		} catch (SQLException e) {
			logger.error("", e);
			throw new RuntimeException(e);
			//TODO better exception
		}
		
		this.getValueSQL = SQL_Translation.sqlGetBufferBlockCounter(program_meta_schema, buffer_id);
		this.incrementValueSQL = SQL_Translation.sqlIncrementBlockCounter(program_meta_schema, buffer_id);
		this.usePreparedStatements = usePreparedStatements;
		
		this.init();
	}
	
	void init(){
		int tryCount = 2;
		
		while(true){
			Connection connection = this.bufferStore.connection();
			try
			{
				this.getValue = SQL_Translation.prepareStatement(connection, this.getValueSQL, this.usePreparedStatements);
				this.incrementValue = SQL_Translation.prepareStatement(connection, this.incrementValueSQL, this.usePreparedStatements);
				break;
			} catch (SQLException f)
			{
				if(SQL_EXCEPTIONS.isConnectionFailure(f)){
					logger.debug("Connection failure. Trying again.", f);
					this.bufferStore.reconnect(connection);
				}
				else if(SQL_EXCEPTIONS.isMonetDBBug(f) & tryCount > 0){
					logger.warn("MonetDB bug. Trying to reconnect.", f);
					this.bufferStore.reconnect(connection);
					--tryCount;
				}
				else{
					logger.error("", f);
					throw new RuntimeException(f);
					//TODO better exception
				}
			}
		}
	}

	@Override
	public synchronized long now() {
		try {
			while(true){
				try{
						ResultSet value;
						value = this.getValue.executeQuery();
						value.next();
						long now = value.getLong(1);
						logger.debug("Retrieved current value of block counter and return now value: " + now);
						try{ this.getValue.getMoreResults(); } catch (SQLException e){}
						return now;
				}
				catch(SQLException e){
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						logger.debug("Connection failure. Trying again.", e);
						this.bufferStore.reconnect(this.getValue.getConnection());
						this.init();
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						logger.debug("Reading block counter rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						logger.debug("Handle of read block counter statement is invalid. Repreparing statement and trying again.", e);
						this.init();
					}
					else{
						throw e;
					}
				}				
			}
		} catch (SQLException e) {
			logger.error("", e);
			throw new RuntimeException(e);
			//TODO better exception
		}
	}
	
	@Override
	public synchronized long min() {
		try {
			while(true){
				try{
					this.incrementValue.executeUpdate();					
					break;
				}
				catch(SQLException e){
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						logger.debug("Connection failure. Trying again.", e);
						this.bufferStore.reconnect(this.getValue.getConnection());
						this.init();
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						logger.debug("Reading block counter rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						logger.debug("Handle of read block counter statement is invalid. Repreparing statement and trying again.", e);
						this.init();
					}
					else{
						throw e;
					}
				}				
			}
			
			//get value
			while(true){
				try{
					ResultSet value;
					value = this.getValue.executeQuery();
					value.next();
					long min = value.getLong(1) - 1;
					logger.debug("Retrieved current value of block counter and return min value (= now-1): " + min);
					try{ this.getValue.getMoreResults(); } catch (SQLException e){}
					return min;
				}
				catch(SQLException e){
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						logger.debug("Connection failure. Trying again.", e);
						this.bufferStore.reconnect(this.getValue.getConnection());
						this.init();
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						logger.debug("Reading block counter rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						logger.debug("Handle of read block counter statement is invalid. Repreparing statement and trying again.", e);
						this.init();
					}
					else{
						throw e;
					}
				}				
			}
		} catch (SQLException e) {
			logger.error("", e);
			throw new RuntimeException(e);
			//TODO better exception
		}
	}

}
